import logging
import os
import pickle
import copy
import json
from typing import Dict, Tuple
from .data_operate import resolve_click_expose_data
from lupin_utils.config import read_config
from lupin_utils.error import ConfigError
from .libsvm import load_libsvm

__all__ = [
    'load_data',
    'load_libsvm'
]

CHECK = ['input', 'features', 'label', 'output']


def load_pickle_data(filepath: str):
    if not os.path.exists(filepath):
        logging.error(f'loading data file error, error file: {filepath}')
        return

    with open(filepath, 'rb') as f:
        res = []
        while True:
            try:
                res.append(pickle.load(f))
            except:
                break
        return res


def load_json_data(filepath: str):
    if not os.path.exists(filepath):
        logging.error(f'loading data file error, error file: {filepath}')
        return

    with open(filepath, 'rb') as f:
        res = json.loads(f.read())
        return res


def _load(filepath: str, param_config: Dict) -> Tuple:
    file_type = param_config['input']['file_type']
    config_features = param_config['features']['names'].replace(' ', '').split(',')
    raw_save_path = param_config['output']['raw_file']
    feature_save_path = param_config['output']['feature_file']
    label_name = param_config['label']['name']
    raw_file = ''
    feature_file = ''
    if file_type == 'pickle':
        res = load_pickle_data(filepath)
        raw_file, feature_file = resolve_click_expose_data(res, config_features, raw_save_path, feature_save_path,
                                                           label_name)
    if file_type.lower() == 'json':
        res = load_json_data(filepath)
        raw_file, feature_file = resolve_click_expose_data(res, config_features, raw_save_path, feature_save_path,
                                                           label_name)
    return raw_file, feature_file


# 加载数据 转换成向量
def load_data(source_data_path: str, config_name: str) -> Tuple:
    """
    :param source_data_path: 数据文件路径
    :param config_name: 配置参数名称
    :return:
    """
    if not os.path.exists(source_data_path):
        raise ConfigError(f'loading data from: {source_data_path}, file not exists')

    param_config = read_config(config_name)

    if len(param_config) == 0:
        raise ConfigError(f'loading data config:{param_config} failed, config is empty')

    check = copy.deepcopy(CHECK)
    # 做一配置参数次检查
    for k in param_config:
        if k in check:
            check.remove(k)
    if len(check) != 0:
        raise ConfigError(f'config missing params: {str(check)}')

    # 从这个地方开始交给子函数处理
    # 加载数据
    raw_file, feature_file = _load(source_data_path, param_config)
    return raw_file, feature_file



# -*- coding:utf-8 -*-
# @author: Henry
# @time: 2019-05-14 10:53 am
# @dec: 使用sklearn PipeLine 做封装
# @WeChat: 13270593018


from sklearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from typing import List, Tuple
from pandas.core.frame import DataFrame


def fit(x_train: DataFrame, y_train: DataFrame, transformers: List, models: List[Tuple]) -> Pipeline:
    """
    训练模型
    :param x_train: 训练样本X
    :param y_train: 训练样本y
    :param transformers: 数据转换器 一般做处理加载数据里的坏值 例如Nan值 不需要处理 transformers=[]
    :param models: 模型
                例如：models = [('classifier', RandomForestClassifier())] 选择随机森林分类器
                如果是包含了多个模型， 则按照pipeline的处理顺序依次处理 例如:
                    models = [('classifier_1', MyClassifier_1), ('classifier_2', MyClassifier_2),....,('classifier_n', MyClassifier_n)]
                则 classifier_1 计算出的结果传递给classifier_2,计算结果依次传递
    :return: 返回pipeline
    """
    steps = []
    if len(transformers) != 0:
        preprocessor = ColumnTransformer(
            transformers=transformers
        )
        steps.append(('preprocessor', preprocessor))
    for m in models:
        steps.append(m)
    rf = Pipeline(steps=steps)

    # 训练模型
    rf.fit(x_train, y_train)

    return rf



# @author: henry
# @time: 2019-05-10
# @dec: 数据编码

from typing import Dict, List
from pandas.io.json import json_normalize
from pandas import DataFrame

__all__ = [
    'encode_feature'
]


def encode_feature(filepath: str, category: str, raw_features: List, start: int) -> (str, Dict, int):
    """
    :param filepath: 原始值写入写入的文件路径
    :param category: 类目
    :param raw_features: 原始特征值
    :param start: 标签编码开始值
    :return: 返回编码的截止值
    """
    i = start
    d = {}
    encoder = []

    for f in raw_features:
        d[f] = i
        encoder.append(i)
        i += 1

    with open(filepath, 'a') as file:
        file.write(f'{category}:' + '\n')
        file.write(f'   {raw_features}' + '\n')
        file.write(f'   {encoder}' + '\n')

    return category, d, i


# 实现点访问
def dict_to_json_normalize(d: Dict, filepath: str = '') -> DataFrame:
    """
    :param d:
    :param filepath: 需要保存成csv的路径
    :return:
    """
    res = json_normalize(d)
    if filepath != '':
        res.to_csv(filepath)
    return res



# -*- coding:utf-8 -*-
# @author: Henry
# @time: 2019-05-17 5:52 pm
# @dec: 处理配置数据
# @WeChat: 13270593018


from lupin_utils.config import read_config
from pandas import DataFrame
from typing import Dict, List


class Preprocessor:
    """
    只用于抽出配置需要的直接数据和间接数据 其中的行列扩展操作留给转换器操作
    """

    def __init__(self, name: str):
        self.direct_features: Dict = {}
        self.indirect_features: Dict = {}
        self.label: str = ''
        self.get_config(name)

    def get_config(self, name: str):
        """
        获取配置文件
        :param name:
        :return:
        """
        configs = read_config(name)
        extraction = configs['extraction']
        self.direct_features = extraction['direct_features']
        self.indirect_features = extraction['indirect_features']
        self.label = configs['label']

    def resolve_direct_features(self, data: DataFrame) -> DataFrame:
        """
        解析直接获取的字段
        :param data:
        :return:
        """
        names = self.direct_features['names'].replace(' ', '').split(',')
        columns = data.columns
        drop_columns = []
        for col_name in columns:
            if col_name not in names:
                drop_columns.append(col_name)

        return data.drop(columns=drop_columns, axis=1)

    def resolve_indirect_features(self, data: DataFrame) -> DataFrame:
        """
        解析间接获取的字段
        :param data:
        :return:
        """
        name = self.indirect_features.get('names', None)
        dtype = self.indirect_features.get('type', None)
        top = self.indirect_features.get('top', None)
        sub_features = self.indirect_features.get('sub_features', None)
        sub_features_type = self.indirect_features.get('sub_features_type', None)
        if sub_features is not None:
            sub_features = sub_features.replace(' ', '').split(',')

        if sub_features_type is not None:
            sub_features_type = sub_features_type.replace(' ', '').split(',')

        extend_dimension_list = []
        print(sub_features)
        count = 0
        for sft in sub_features_type:
            if sft.lower() == 'list':
                extend_dimension_list.append(sub_features[count])
            count += 1

        drop_data = data[name].to_frame()
        set_columns = [f'{name}.{sub_name}' for sub_name in sub_features]
        print(set_columns)
        print(extend_dimension_list)

        return drop_data



import yaml
import os

config_path = os.path.abspath(os.path.join(__file__, '../../config/config.yaml'))


def read_key(key: str):
    print(config_path)
    config = yaml.load(open(config_path, 'r', encoding="utf-8"))
    res = config.get(key, None)
    if res is None:
        raise KeyError(f'{key} not in config.yaml')
    return res



# -*- coding:utf-8 -*-
# @author: Henry
# @time: 2019-05-17 11:28 am
# @dec: 加载原始文件数据 原始文件数据类型json
# @WeChat: 13270593018

import os
import pickle
import json
from lupin_utils.error import FileNotExistsError
from pandas.io.json import json_normalize
from pandas import DataFrame

__all__ = [
    'FileOperate'
]


class FileOperate:
    def __init__(self):
        pass

    @staticmethod
    def transform_pickle_as_json(filepath: str, savepath: str) -> json:
        if not os.path.exists(filepath):
            raise FileNotExistsError(f'pickle file not exists, file path:{filepath}')

        res = []
        with open(filepath, 'rb') as fp:
            while True:
                try:
                    res.append(pickle.load(fp))
                except:
                    break

        with open(savepath, 'w') as fj:
            fj.write(json.dumps(res))

    @staticmethod
    def load_json(filepath: str) -> json:
        if not os.path.exists(filepath):
            raise FileNotExistsError(f'json file not exists, file path:{filepath}')

        with open(filepath, 'rb') as file:
            return json.loads(file.read())

    @staticmethod
    def normalize_json(data: json) -> DataFrame:
        try:
            return json_normalize(data)
        except Exception as e:
            raise e



# -*- coding:utf-8 -*-
# @author: Henry
# @time: 2019-05-14 09:44 am
# @dec: 加载libsvm格式数据
# @WeChat: 13270593018

from sklearn.datasets import load_svmlight_file
from sklearn.model_selection import train_test_split
from typing import Tuple
from lupin_utils.error import FileNotExistsError
import os
import logging
import pandas as pd


def load_libsvm(filepath: str, test_size: float = 0.25, random_state=None) -> Tuple:
    """
    加载libsvm 数据并且进行转换
    :param filepath: libsvm 数据文件
    :param test_size: 测试数据占比
    :param random_state: int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.
    :return: 如果float 大于零 则返回长度为4的Tuple (X_train, X_test, y_train, y_test) 否则返回长度为2Tuple (X, y)
    """
    if not os.path.exists(filepath):
        logging.error(f'file - {filepath} - not exists')
        raise FileNotExistsError()

    x_load, y_load = load_svmlight_file(filepath)
    x = pd.DataFrame(x_load.todense())
    y = pd.DataFrame(y_load)
    if test_size > 0.0:
        x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=test_size, random_state=random_state)
        return x_train, x_test, y_train, y_test

    return x, y



# -*- coding:utf-8 -*-
# @author: Henry
# @time: 2019-05-15 8:35 pm
# @dec: 封装数据处理
# @WeChat: 13270593018

from pandas import DataFrame
from typing import List, Tuple


class JoinData:
    """
    data frame 数据join
    """

    @staticmethod
    def left_join(x: DataFrame, y: DataFrame) -> DataFrame:
        """
        左连接
        :param x:
        :param y:
        :return:
        """
        return x.join(y)

    @staticmethod
    def merge(x: DataFrame, y: DataFrame) -> DataFrame:
        """
        merge
        :param x:
        :param y:
        :return:
        """
        return x.join(y, how='outer')


class BaseTransformer:
    """
    数据预处理transformer基类
    在派生类里面需要完成fit的实现
    """

    def __init__(self, *args, **kwargs):
        """
        数据预处理模型参数选择
        :param args:
        """
        self.args = args
        self.kwargs = kwargs

    def fit(self, x: DataFrame, y: DataFrame) -> (DataFrame, DataFrame):
        pass


class PipeDataTransformer:
    """
       类似于PipeLine， 用于数据预处理
       """

    def __init__(self, steps: List[Tuple[str, BaseTransformer]]):
        self.steps = steps

    def fit(self, x_train: DataFrame, y_train: DataFrame) -> (DataFrame, DataFrame):
        """
        按steps顺序依次处理并传递数据
        :param x_train:
        :param y_train:
        :return:
        """
        x = x_train
        y = y_train
        for step in self.steps:
            transformer = step[1]
            x, y = transformer.fit(x, y)

        return x, y



# -*- coding:utf-8 -*-
# @author: Henry
# @time: 2019-05-16 2:27 pm
# @dec: multiple columns label-encoder and one-hot-encoder
# @WeChat: 13270593018

from .base import BaseTransformer
from pandas import DataFrame
from typing import List, Tuple
from lupin_utils.error import ParamTypeError
from sklearn.preprocessing import LabelEncoder, OneHotEncoder


class MultipleColumnLabelEncoder(BaseTransformer):
    """
    多维度特征label编码
    """

    def __init__(self, *args, **kwargs):
        BaseTransformer.__init__(self, *args, **kwargs)

    def fit(self, x: DataFrame, y: DataFrame) -> (DataFrame, DataFrame):
        output = x.copy()
        if 'columns' in self.kwargs:
            columns = self.kwargs['columns']
            param_type = type(columns)
            if param_type not in [List, Tuple]:
                raise ParamTypeError(
                    f'class: {MultipleColumnLabelEncoder.__name__} needs param type: {List} or {Tuple} but got {param_type}')

            for col in columns:
                output[col] = LabelEncoder().fit_transform(output[col])

        else:
            for col_name, col in output.iteritems():
                output[col_name] = LabelEncoder().fit_transform(col)

        return output, y


class LupinOneHotEncoder(BaseTransformer):
    """
    one-hot encode
    """

    def __init__(self, *args, **kwargs):
        BaseTransformer.__init__(self, *args, **kwargs)
        self.args = args

    def fit(self, x: DataFrame, y: DataFrame) -> (DataFrame, DataFrame):
        return OneHotEncoder(*self.args, **self.kwargs).fit_transform(x), y



# -*- coding:utf-8 -*-
# @author: Henry
# @time: 2019-05-13 2:52 pm
# @dec: 预处理工具
# @file: lupin_utils


"""
    1、标准化
    2、最小-最大规范化
    3、规范化
    4、特征二值化
"""

import json
from sklearn.preprocessing import scale, normalize, Binarizer
from numpy import array
from pandas.io.json import json_normalize
from pandas import DataFrame

__all__ = [
    'standardization',
    'normalization',
    'binarizer',
    'normalize_json'
]


def standardization(data) -> array:
    """
    标准化
    :param data:
    :return:
    """
    return scale(data)


def normalization(data, norm: str = 'l2') -> array:
    """
    规范化
    :param data:
    :param norm: 默认l2
    :return:
    """
    return normalize(data, norm)


def binarizer(data) -> array:
    """
    特征二值化
    :param data:
    :return:
    """
    return Binarizer(data)


def normalize_json(data: json) -> DataFrame:
    """
    返回平坦化格式数据 通过访问列来获取所有数据
    :param data:
    :return:
    """
    return json_normalize(data)



import datetime
from zipfile import ZipFile


def show(t: datetime.datetime = datetime.datetime.now()):
    print(t.strftime("%Y%m%d%H%M%S"))


def test_zip(fp: str, sp: str):
    with ZipFile(fp, 'r') as myzip:
        myzip.write(sp)


if __name__ == '__main__':
    # show()
    fp = '/Users/henry/PycharmProjects/lupin_utils/models/best_estimator'
    sp = '/Users/henry/PycharmProjects/lupin_utils/model/best_estimator'
    test_zip(fp, sp)



# -*- coding:utf-8 -*-
# @author: Henry
# @time: 2019-05-14 4:40 pm
# @dec: 保存训练好的模型
# @WeChat:13270593018

import os
from sklearn.externals import joblib
from sklearn.pipeline import Pipeline

FILE_ROOT = os.getcwd()

MODEL_PATH = os.path.abspath(os.path.join(FILE_ROOT, 'models'))


def save_training_model(model: Pipeline, model_name: str):
    """
    保存训练的模型
    :param model:
    :param model_name:
    :return:
    """
    file_path = os.path.join(MODEL_PATH, model_name)
    joblib.dump(model, file_path, compress=1)



from typing import List, Tuple
from .operate import label_and_one_hot_encode
import gc


def resolve_click_expose_data(data: List, config_features: List, raw_save_path: str,
                              feature_save_path: str, label_name: str
                              ) -> Tuple:
    """
    保存格式: label1 feature1:value1 feature2:value2 feature3:value3 ......
    :param data:  读取的原始的值 这个是一个list 每个元素里面有相应的dict 存放需要读取的特征
    :param config_features: 配置里需要提取的特征
    :param raw_save_path: 原始数据存储文件名
    :param feature_save_path: 特征编码转换存储的位置
    :param label_name: label 字段
    :return:
    """
    # 提取所有feature 按类目配置提取的字段提取 存储方式 [set(), set(), set()] 每一个set() 对应一个需要提取的配置的特征
    features = {}
    for f in config_features:
        features[f'{f}'] = set()

    for item in data:
        for k, v in item.items():
            if k == 'response_body':
                for target in v['target']:
                    # 遍历指定feature
                    for f in config_features:
                        try:
                            for sub_feature in target[0][f]:
                                features[f].add(sub_feature)
                        except:
                            continue

    for k, v in features.items(): features[k] = list(v)

    label_encoder = label_and_one_hot_encode(features, raw_save_path)
    check_label = []
    with open(feature_save_path, 'w') as file:
        file.write('')
    for item in data:
        for k, v in item.items():
            if k == 'response_body':
                for target in v['target']:
                    try:
                        # 提取label 这里为spu
                        label = target[0][label_name]
                        if label not in check_label:
                            check_label.append(label)
                            encoder_list = []
                            for fea in config_features:
                                sub_feature = [f'{fea}.{x}' for x in list(set(target[0][fea]))]
                                encoder_list += list(label_encoder.transform(sub_feature))
                            sorted_encoder = sorted(encoder_list)
                            with open(feature_save_path, 'a+') as file:
                                if label < 200000:
                                    line = f'{0} '
                                else:
                                    line = f'{1} '
                                # line = f'{label} '
                                for i in sorted_encoder:
                                    line = line + f' {i}:{1}'
                                file.write(line + '\n')
                        else:
                            pass
                    except:
                        continue

    gc.collect()

    return raw_save_path, feature_save_path



# -*- coding:utf-8 -*-
# @author: Henry
# @time: unknown
# @dec: 模型评估
# @WeChat: 13270593018
# @modify: 2019-05-15

from sklearn.metrics import accuracy_score
from sklearn.metrics import cohen_kappa_score
from sklearn.metrics import confusion_matrix
from sklearn.metrics import hamming_loss
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from pandas import DataFrame
from numpy import ndarray

__all__ = [
    'score_accuracy',
    'score_cohen_kappa',
    'score_hamming_loss',
    'calculate_confusion_matrix'
]


def score_accuracy(y_true: DataFrame, y_pred: DataFrame) -> float:
    """
    计算准确率
    :param y_true:
    :param y_pred:
    :return:
    """
    return accuracy_score(y_true, y_pred)


def score_cohen_kappa(y_true: DataFrame, y_pred: DataFrame) -> float:
    """
    计算 Kappa系数 衡量数据的一致性
    :param y_true:
    :param y_pred:
    :return:
    """
    return cohen_kappa_score(y_true, y_pred)


def score_hamming_loss(y_true: DataFrame, y_pred: DataFrame) -> float:
    """
    计算hamming loss
    :param y_true:
    :param y_pred:
    :return:
    """
    return hamming_loss(y_true, y_pred)


def calculate_confusion_matrix(y_true: DataFrame, y_pred: DataFrame) -> ndarray:
    """
    计算混淆矩阵
    :param y_true:
    :param y_pred:
    :return:
    """
    return confusion_matrix(y_true, y_pred)


def score_precision(y_true: DataFrame, y_pred: DataFrame) -> float:
    """
    计算准确率
    :param y_true:
    :param y_pred:
    :return:
    """
    return precision_score(y_true, y_pred)


def score_recall(y_true: DataFrame, y_pred: DataFrame) -> float:
    """
    计算召回率
    :param y_true:
    :param y_pred:
    :return:
    """
    return recall_score(y_true, y_pred)



# -*- coding:utf-8 -*-
# @author: Henry
# @time: 2019-05-16 10:24 am
# @dec: 多特征标签编码
# @WeChat: 13270593018

from sklearn.preprocessing import LabelEncoder, OneHotEncoder


class MultiColumnEncoder:
    def __init__(self, columns=None):
        self.columns = columns

    def transform(self, x):
        output = x.copy()
        if self.columns is not None:
            for col in self.columns:
                output[col] = LabelEncoder().fit_transform(output[col])
        else:
            for col_name, col in output.iteritems():
                output[col_name] = LabelEncoder().fit_transform(col)

        return output

    def fit_transform(self, x):
        return self.transform(x)



# -*- coding:utf-8 -*-
# @author: Henry
# @time: 2019-05-20 10:23 pm
# @dec: dataframe 连接
# @WeChat: 13270593018

import json
from pandas import DataFrame
from typing import List, Tuple, Dict
from .base import JoinData
from sklearn.preprocessing import LabelEncoder


def get_index(idx: int, value: List[int]) -> int:
    """
    计算偏移量
    :param idx:
    :param value:
    :return:
    """
    offset = 0
    if idx == 0:
        return idx
    for i in range(0, idx - 1):
        offset += value[i]

    return offset


def calculate_range_label(value: float, range_array: List):
    for i in range_array:
        if i < 0:
            raise Exception('range label exits value negative')
    label = 0
    for i in range_array:
        if value < i:
            return label
        else:
            label += 1

    return label


def file_write(fp: str, way: str, data: str):
    with open(fp, way) as f:
        if data == '':
            f.write(data)
        else:
            f.write(data + '\n')


class Extraction:
    """
   提取源文件
    """

    def __init__(self, direct_config_features: List, indirect_config_features: List, indirect_config_sub_features: List,
                 indirect_config_sub_features_type: List, top: int, positive: str, others: Dict):
        self.direct_features_dict = {}
        self.indirect_features_dict = {}
        self.direct_config_features = direct_config_features
        self.indirect_config_features = indirect_config_features
        self.indirect_config_sub_features = indirect_config_sub_features
        self.indirect_config_sub_features_type = indirect_config_sub_features_type
        self.top = top
        self.others = others
        self.multiple_fields = []
        self.single_fields = []
        self.continuous_fields = []
        self.positive = positive

    @staticmethod
    def extract__(raw_data: DataFrame, config_features: List, raw_path: str, coded_path: str):
        """
        抽取字段
        :param raw_data:
        :param config_features
        :param raw_path:
        :param coded_path:
        :return:
        """
        columns = raw_data.columns
        # 去除不需要的列元素
        drop_columns = []
        for name in columns:
            if name not in config_features:
                drop_columns.append(name)

        # 保存计算的值
        data = raw_data.drop(columns=drop_columns, axis=1)
        columns = data.columns
        # 排序一次
        columns = sorted(columns)
        # 设置特征映字典
        features = {col: set() for col in columns}
        for col_name in columns:
            col = [str(item) for item in data[col_name]]
            for c in col:
                features[col_name].add(c)

        # 对多维特征进行一次排序
        features = {k: sorted(list(features[k])) for k in features}

        # 文件初始化为空
        file_write(raw_path, 'w', '')
        file_write(coded_path, 'w', '')

        # 写入原始数据 该数据存入顺序为 配置提取字段的顺序
        for seg in config_features:
            file_write(raw_path, 'a', f'{seg}:')
            file_write(raw_path, 'a', f'{str(features[seg])}')

        # 编码值 行循环计算 这里索引值就是编码值
        for _, row in data.iterrows():
            index_lable = []
            # 拿到每一个特征编码值反编码
            for col in config_features:
                value = data[col]
                code_value = features[col].index(value)
                index_lable.append(code_value)

            line = ''
            for i in range(0, len(index_lable)):
                v = index_lable[i]
                line += f'{i}:{v} '

            file_write(coded_path, 'a', line)

    def direct_extract(self, raw_data: DataFrame):
        """
        抽取原始值并且编码 存储方式为dict
        :param raw_data:
        :return:
        """
        config_features = self.direct_config_features
        columns = raw_data.columns
        # 去除不需要的列元素
        drop_columns = []
        for name in columns:
            if name not in config_features:
                drop_columns.append(name)

        # 保存计算的值
        data = raw_data.drop(columns=drop_columns, axis=1)
        columns = data.columns
        # 设置特征集合
        features = {col: {} for col in columns}
        columns = data.columns
        for col in columns:
            cols = list(data[col])
            label_code = LabelEncoder().fit_transform(cols)
            for i in range(0, len(cols)):
                features[col][cols[i]] = int(label_code[i])

        self.direct_features_dict = features
        # # 文件初始化为空
        # file_write(raw_path, 'w', '')
        # file_write(coded_path, 'w', '')
        # file_write(raw_path, 'a', json.dumps(features))

    @staticmethod
    def split_dimension(config_features: List, config_types: List) -> Tuple:
        """
        拆分单维特征和多维特征的字段以及连续字段
        :return:
        """

        multiple_fields = []
        single_fields = []
        continuous_fields = []

        if len(config_features) != len(config_types):
            raise Exception('length not matches')
        for i in range(0, len(config_features)):
            cn = config_features[i]
            cnt = config_types[i]
            if cnt.lower() == 'list':
                multiple_fields.append(cn)
            elif cnt.lower() == 'continuous':
                continuous_fields.append(cn)
            else:
                single_fields.append(cn)

        return multiple_fields, single_fields, continuous_fields

    def indirect_extract(self, raw_data: DataFrame):
        """
        间接提取的字段值 目前只能为单个字段 使用时用List表示
        提取单个
        :return:
        """
        segments = self.indirect_config_features
        config_features = self.indirect_config_sub_features
        config_type = self.indirect_config_sub_features_type
        top = self.top

        prefix_feature = segments[0]

        columns = raw_data.columns
        # 去除不需要的列元素
        drop_columns = []
        for name in columns:
            if name not in segments:
                drop_columns.append(name)

        # 保存计算的值
        data = raw_data.drop(columns=drop_columns, axis=1)

        # 抽取出单维度字段和多维度字段
        multiple_fields, single_fields, continuous_fields = self.split_dimension(config_features, config_type)

        self.multiple_fields = multiple_fields
        self.single_fields = single_fields
        self.continuous_fields = continuous_fields

        # 给所有特征编码 并且存储
        single_features = {sf: set() for sf in single_fields}
        multiple_features = {mf: set() for mf in multiple_fields}
        for item in data[prefix_feature]:
            top_data = item[0:top]
            # 遍历前top每一条数据
            for td in top_data:
                # 获取每一个single_field 值
                for s in single_fields:
                    value = td[0][s]
                    single_features[s].add(value)

                for m in multiple_fields:
                    value_list = td[0][m]
                    for f in value_list:
                        multiple_features[m].add(f)

        # 处理单维度特征
        single_feature_dict = {sf: {} for sf in single_fields}
        for f in single_features:
            raw = list(single_features[f])
            label_code = LabelEncoder().fit_transform(raw)
            for i in range(0, len(raw)):
                k = raw[i]
                v = label_code[i]
                single_feature_dict[f][k] = v

        # 给multiple_features 做一次值处理 主要是为了处理列位置
        for k, v in multiple_features.items():
            multiple_features[k] = {}
            for i, element in enumerate(v):
                multiple_features[k][element] = i

        indirect_features_dict = {prefix_feature: {}}
        for k in single_feature_dict:
            indirect_features_dict[prefix_feature][k] = single_feature_dict[k]

        for k in multiple_features:
            indirect_features_dict[prefix_feature][k] = multiple_features[k]

        self.indirect_features_dict = indirect_features_dict

    def layout(self):
        """
        排版libsvm的排列格式
        :return:
        """
        column_range = []
        column_range += self.direct_config_features
        for f in self.single_fields:
            column_range.append(f'{self.indirect_config_features[0]}.{f}')

        print(column_range)

    def save_raw(self, raw_path: str):
        """
        保存原始值以及其对应关系
        顺序依次为 直接提取参数顺序 + 加间接提取单维度字段 + 间接提取连续字段 + 间接提取多维度字段
        其中各个配置顺序按给定的配置顺序编排

        :return:
        :return:
        """
        file_write(raw_path, 'w', '')
        columns = self.direct_config_features + [f'{self.indirect_config_features[0]}.{i}' for i in
                                                 self.single_fields + self.continuous_fields + self.multiple_fields]
        file_write(raw_path, 'a', json.dumps(columns))

        file_write(raw_path, 'a', json.dumps(self.direct_features_dict))

        d = {}
        fixed_prefix = self.indirect_config_features[0]
        d[fixed_prefix] = {}
        for k in self.multiple_fields:
            d[fixed_prefix][k] = self.indirect_features_dict[self.indirect_config_features[0]][k]

        file_write(raw_path, 'a', json.dumps(d))

    def extract(self, raw_data: DataFrame, code_path: str):
        file_write(code_path, 'w', '')
        drop_columns = []
        columns = raw_data.columns
        save_columns = self.direct_config_features + self.indirect_config_features
        for col in columns:
            if col not in save_columns:
                drop_columns.append(col)

        data = raw_data.drop(columns=drop_columns, axis=1)

        for _, row in data.iterrows():
            direct_value = []
            # 提取直接字段特征
            for col_name in self.direct_config_features:
                # 如果是直接提取的数据
                value = row[col_name]
                direct_value.append(self.direct_features_dict[col_name][value])

                # print(direct_dict)
            # 该字段只有一个
            indirect_col_name = self.indirect_config_features[0]
            indirect_value = row[indirect_col_name]
            top_data = indirect_value[0:self.top]
            # 计算实际取到的长度 用于行扩展
            for item in top_data:
                # 单维度处理
                single_value = []
                for index_name in self.single_fields:
                    # 拿到编码值
                    value = item[0][index_name]
                    single_value.append(self.indirect_features_dict[indirect_col_name][index_name][value])

                # 连续数据处理
                continuous_value = []
                for index_name in self.continuous_fields:
                    value = item[0][index_name]
                    discrete_rule = self.others[index_name]
                    if index_name == 'sales_stat':
                        encode_value = calculate_range_label(float(value['30']['salesCount']), discrete_rule)
                    else:
                        encode_value = calculate_range_label(float(value), discrete_rule)
                    continuous_value.append(encode_value)

                single_line = ''
                to_file = direct_value + single_value + continuous_value
                for i in range(0, len(to_file)):
                    single_line += f'{i}:{to_file[i]} '

                # 多维度处理
                tmp_length = len(to_file)
                multiple_line = ''
                for index_name in self.multiple_fields:
                    value = item[0][index_name]
                    value = set(value)
                    label = []
                    for tag in value:
                        # 计算出编码值
                        label.append(self.indirect_features_dict[indirect_col_name][index_name][tag])
                    label = sorted(label)
                    for l in label:
                        multiple_line += f'{l + tmp_length}:{1} '

                    tmp_length += len(self.indirect_features_dict[indirect_col_name][index_name])
                line = single_line + multiple_line
                # 标签输出值
                if item[1] == self.positive:
                    line = f'{1}  ' + line
                else:
                    line = f'{0}  ' + line

                file_write(code_path, 'a', line)



# -*- coding:utf-8 -*-
# @author: Henry
# @time: None
# @dec: Error Class
# @WeChat: 13270593018

__all__ = [
    'FileNotExistsError',
    'ConfigError',
    'MlflowConfigError',
    'ParamTypeError'
]


class LupinException(Exception):
    def __init__(self, *args):
        self.args = args


class FileNotExistsError(LupinException):
    def __init__(self, *args):
        self.args = args
        self.message = '文件不存在'
        self.code = 1


class ConfigError(LupinException):
    def __init__(self, args):
        self.args = args
        self.message = 'yaml 配置错误'
        self.code = 2


class MlflowConfigError(LupinException):
    def __init__(self, *args):
        self.args = args
        self.message = 'mlflow 配置错误'
        self.code = 3


class ParamTypeError(LupinException):
    def __init__(self, *args):
        self.args = args
        self.message = '参数类型错误'
        self.code = 4



# -*- coding:utf-8 -*-
# @author: Henry
# @time: 2019-05-15 8:35 pm
# @dec: 封装数据处理
# @WeChat: 13270593018

from pandas import DataFrame
from typing import List, Tuple


class BaseTransformer:
    """
    数据预处理transformer基类
    在派生类里面需要完成fit的实现
    """

    def __init__(self, *args):
        """
        数据预处理模型参数选择
        :param args:
        """
        self.args = args

    def fit(self, x: DataFrame, y: DataFrame) -> (DataFrame, DataFrame):
        pass


class TransformerOperate:
    """
    类似于PipeLine， 用于数据预处理
    """

    def __init__(self, steps=List[Tuple[str, BaseTransformer()]]):
        self.steps = steps

    def fit(self, x_train: DataFrame, y_train: DataFrame) -> (DataFrame, DataFrame):
        x = x_train
        y = y_train
        for step in self.steps:
            transformer = step[1]
            x, y = transformer.fit(x, y)

        return x, y



# -*- coding:utf-8 -*-
# @author: Henry
# @time: 2019-05-15 9:48 am
# @dec: mlflow封装
# @WeChat: 13270593018
# @requirements: mlflow==0.8.2

import mlflow
import mlflow.sklearn
from sklearn.pipeline import Pipeline
from lupin_utils.error import MlflowConfigError

__all__ = [
    'LupinMlflow'
]


class LupinMlflow:
    def __init__(self, tracing_host=None, tracing_port=None, use_https=False):
        """
        初始化远程服务地址
        :param tracing_host: 远程 mlflow host
        :param tracing_port: 远程 mlflow port
        """
        self.tracing_host = tracing_host
        self.tracing_port = tracing_port
        self.mlflow = mlflow
        self.use_https = use_https
        self.tracing_init()

    def tracing_init(self):
        if self.tracing_host is None or self.tracing_port is None:
            raise MlflowConfigError(
                f'tracing url missing right config, host:{self.tracing_host}, port:{self.tracing_port}')
        if self.use_https:
            url = f'https://{self.tracing_host}:{self.tracing_port}'
        else:
            url = f'http://{self.tracing_host}:{self.tracing_port}'
        self.mlflow.set_tracking_uri(url)

    def log_model(self, model: Pipeline, name: str):
        """
        上传训练模型
        :param model:
        :param name:
        :return:
        """
        self.mlflow.sklearn.log_model(model, name)

    def log_metrics(self, **kwargs):
        """
        上传计算结果
        :param kwargs:
        :return:
        """
        for k, v in kwargs.items():
            self.mlflow.log_metric(k, v)

    def log_params(self, **kwargs):
        """
        上传参数
        :param kwargs:
        :return:
        """
        for k, v in kwargs.items():
            self.mlflow.log_param(k, v)



# -*- coding:utf-8 -*-
# @author: Henry
# @time: 2019-05-13 10:45 am
# @dec: one-hot encode
# @file: lupin_utils
# @WeChat: 13270593018

from typing import List, Dict
from sklearn.preprocessing import LabelEncoder
import json


def label_and_one_hot_encode(data: Dict, raw_file: str) -> LabelEncoder:
    """
    :param data:  k, v（str，List）分别为 需要提取的字段和对应的特征列表
    :param raw_file 原始值保存的位置
    :return:
    """
    raw = []
    for k, v in data.items():
        for fea in v:
            raw.append(f'{k}.{fea}')

    label_encoder = LabelEncoder()
    labels = label_encoder.fit_transform(raw)

    # 做一次排序
    labels = list(labels)
    length = len(labels)

    raw_range = []

    for i in range(length):
        index = labels.index(i)
        raw_range.append(raw[index])

    # save
    with open(raw_file, 'w') as file:
        json_data = json.dumps(raw_range)
        file.write(json_data)

    return label_encoder
